! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: esrcfetch_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Fetches information from the eboxdetect source list and writes it to the emldetect list (or vice versa).
! 
! TODO:
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module esrcfetch_mod

  use dal
  use param  
  use errorHandling

  use array_utils	! for smallestTrueIndex, findUniqueValues
  use dal_aux		! for splitSetTabName, readColDataToPtr, addOrOpenColumn

  implicit none 
  
  private 

  ! Command-line parameters:
  character(256) ::&
    inSetTabName = '',&
    outSetTabName = ''
  character(40) ::&
    inIdSrcColName = '',&
    outIdSrcColName = '',&
    flagColName=''
  character(40), allocatable ::&
    infoColNames(:)
  logical ::&
    flagSplitSources=.false.,&
    withUserIntNull=.false.
  integer ::&
    flagBit=0,&
    userIntNullValue=0

  public esrcfetch

contains 
  
  !---------------------------------------------------------------------
  subroutine esrcfetch
  !---------------------------------------------------------------------

    implicit none

    character(256) :: inSetName, outSetName
    character(20) :: inTabName, outTabName
    type(DataSetT) :: inSet, outSet
    type(TableT) :: inTab, outTab
    integer ::&
      numInRows, numOutRows, row, uniqueIndex, i, infoColI
    integer(int32), pointer, dimension(:) ::&
      inIdSrcData, inIdInstData, inIdBandData, outIdSrcData, outIdInstData,&
      outIdBandData, outIdInSrcData, outFlag, uniqueInIdSrcs,&
      flagDataInt32, inInfoDataInt32, outInfoDataInt32
    integer, allocatable :: outToIn(:)
    logical, allocatable :: outNulls(:), mask(:), splitInIdSrcs(:)
    type(ColumnT) :: flagCol, inInfoCol, outInfoCol
    integer(int32) :: flagDataType, intNullValue, inDataType
    integer(int8), pointer ::&
      flagDataInt8(:), inInfoDataInt8(:), outInfoDataInt8(:)
    integer(int16), pointer ::&
      flagDataInt16(:), inInfoDataInt16(:), outInfoDataInt16(:)
    real(single), pointer :: inInfoDataReal32(:), outInfoDataReal32(:)
    real(double), pointer :: inInfoDataReal64(:), outInfoDataReal64(:)

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    call splitSetTabName(inSetTabName, inSetName, inTabName)
    call splitSetTabName(outSetTabName, outSetName, outTabName)
    ! splitSetTabName is in ssclib/dal_aux.

    inSet = dataSet(inSetName, READ)		! eg eboxdetect list
    inTab = table(inSet, inTabName)
    numInRows = numberOfRows(inTab)

    call readColDataToPtr(inTab, inIdSrcColName, inIdSrcData) ! eg BOX_ID_SRC
    call readColDataToPtr(inTab, 'ID_INST',      inIdInstData)
    call readColDataToPtr(inTab, 'ID_BAND',      inIdBandData)
    ! readColDataToPtr is in ssclib/dal_aux.

    outSet = dataSet(outSetName, MODIFY)	! eg emldetect list
    outTab = table(outSet, outTabName)
    numOutRows = numberOfRows(outTab)

    call readColDataToPtr(outTab, outIdSrcColName, outIdSrcData) ! eg ML_ID_SRC
    call readColDataToPtr(outTab, 'ID_INST',       outIdInstData)
    call readColDataToPtr(outTab, 'ID_BAND',       outIdBandData)
    call readColDataToPtr(outTab, inIdSrcColName,  outIdInSrcData)
    ! eg BOX_ID_SRC
    ! readColDataToPtr is in ssclib/dal_aux.

    !...................................................................
    ! Make column-to-column map:

    allocate(outToIn(0:numOutRows-1), outNulls(0:numOutRows-1))
    allocate(mask(numOutRows))
    outToIn = -1 ! defaults to this illegal value.
    outNulls = .false. ! default
    do row = 0, numOutRows-1
      mask = ( inIdSrcData == outIdInSrcData(row)&
        .and. inIdInstData == outIdInstData(row)&
        .and. inIdBandData == outIdBandData(row))

      ! There should be exactly 1 true mask element. Why? Well, if there
      ! are 0, this means that the out (emldetect) table contains a
      ! reference to an in (eboxdetect) source+inst+band combination
      ! which however is not present in the in (eboxdetect) list. We'll
      ! allow this and just write a null under these circumstances. If
      ! there are >1, this means that the in (eboxdetect) list has more
      ! than 1 of the given source+inst+band combination. This is illegal.

if (count(mask) > 1) call error('',"")
      if (count(mask) < 1) then
        outNulls(row) = .true.
    cycle
      end if

      outToIn(row) = smallestTrueIndex(mask) - 1
      ! smallestTrueIndex is in ssclib/array_utils.
    end do
    deallocate(mask)

    !...................................................................

    if (flagSplitSources) then
      if (hasColumn(outTab, flagColName)) then
        flagCol = column(outTab, flagColName)
        flagDataType = dataType(flagCol)
if (.not.(dataTypeIsInteger(flagDataType))) call error('',"")
        call readColDataToPtr(flagCol, outFlag)

      else
        flagDataType = INTEGER32
        flagCol = addColumn(outTab, flagColName, flagDataType)
        call readColDataToPtr(flagCol, outFlag)
        outFlag = 0
      end if

      uniqueInIdSrcs => findUniqueValues(inIdSrcData)
      ! findUniqueValues is in ssclib/array_utils.

      allocate(splitInIdSrcs(size(uniqueInIdSrcs)))
      splitInIdSrcs = .false. ! default

      allocate(mask(numInRows))
      do row = 0, numInRows-1
        mask = (outIdInSrcData == inIdSrcData(row)&
          .and.  outIdInstData == inIdInstData(row)&
          .and.  outIdBandData == inIdBandData(row))

        ! Usually one would expect that there would be just 1 true
        ! element of mask. 0 true elements means the in (eboxdetect)
        ! source isn't found in the out (emldetect) list. Warn in this
        ! case. >1 true elements indicates that the in (eboxdetect)
        ! source is present more than once in the out (emldetect) list.

        if (count(mask) < 1) then
          call warning('missingSource',"")
        else if (count(mask) > 1) then
          uniqueIndex = smallestTrueIndex(uniqueInIdSrcs == inIdSrcData(row))
          splitInIdSrcs(uniqueIndex) = .true.
        end if
      end do

      if (any(splitInIdSrcs)) then
        do i = 1, size(uniqueInIdSrcs)
        if (.not.splitInIdSrcs(i)) cycle
          where(outIdInSrcData == uniqueInIdSrcs(i))&
            outFlag = ibset(outFlag, flagBit)
        end do
      end if

      deallocate(mask, splitInIdSrcs)
      deallocate(uniqueInIdSrcs)

      select case(flagDataType)
      case (INTEGER8)
        flagDataInt8 => int8Data(flagCol)
        flagDataInt8 = outFlag
      case (INTEGER16)
        flagDataInt16 => int16Data(flagCol)
        flagDataInt16 = outFlag
      case (INTEGER32)
        flagDataInt32 => int32Data(flagCol)
        flagDataInt32 = outFlag
      case default
call error('bug',"")
      end select

      deallocate(outFlag)
      call release(flagCol)
    end if

    !...................................................................
    ! Now copy the info over:

    if (withUserIntNull) then
      intNullValue = userIntNullValue
    else
      intNullValue = -99
    end if

    do infoColI = 1, size(infoColNames)
      inInfoCol = column(inTab, infoColNames(infoColI))
      inDataType = dataType(inInfoCol)
      outInfoCol = addOrOpenColumn(outTab, infoColNames(infoColI)&
        , inDataType, units(inInfoCol))!!!!, comment(inInfoCol))
      ! addOrOpenColumn is in ssclib/dal_aux.

      select case(inDataType)
      case(INTEGER8)
        inInfoDataInt8  => int8Data(inInfoCol)
        outInfoDataInt8 => int8Data(outInfoCol)

        call setNullValue(outInfoCol, int(int(intNullValue, int8), int32))

        do row = 0, numOutRows-1
          if (outNulls(row)) then
            call setToNull(outInfoCol, row)
          else
            outInfoDataInt8(row) = inInfoDataInt8(outToIn(row))
          end if
        end do

      case(INTEGER16)
        inInfoDataInt16  => int16Data(inInfoCol)
        outInfoDataInt16 => int16Data(outInfoCol)

        call setNullValue(outInfoCol, int(int(intNullValue, int16), int32))

        do row = 0, numOutRows-1
          if (outNulls(row)) then
            call setToNull(outInfoCol, row)
          else
            outInfoDataInt16(row) = inInfoDataInt16(outToIn(row))
          end if
        end do

      case(INTEGER32)
        inInfoDataInt32  => int32Data(inInfoCol)
        outInfoDataInt32 => int32Data(outInfoCol)

        call setNullValue(outInfoCol, intNullValue)

        do row = 0, numOutRows-1
          if (outNulls(row)) then
            call setToNull(outInfoCol, row)
          else
            outInfoDataInt32(row) = inInfoDataInt32(outToIn(row))
          end if
        end do

      case(REAL32)
        inInfoDataReal32  => real32Data(inInfoCol)
        outInfoDataReal32 => real32Data(outInfoCol)

        do row = 0, numOutRows-1
          if (outNulls(row)) then
            call setToNull(outInfoCol, row)
          else
            outInfoDataReal32(row) = inInfoDataReal32(outToIn(row))
          end if
        end do

      case(REAL64)
        inInfoDataReal64  => real64Data(inInfoCol)
        outInfoDataReal64 => real64Data(outInfoCol)

        do row = 0, numOutRows-1
          if (outNulls(row)) then
            call setToNull(outInfoCol, row)
          else
            outInfoDataReal64(row) = inInfoDataReal64(outToIn(row))
          end if
        end do

      case(STRING)
call error('notYetSupported',"")
      case(BOOLEAN)
call error('notYetSupported',"")
      case default
call error('',"")
      end select

      call release(outInfoCol)
      call release(inInfoCol)
    end do ! loop over info cols

    !...................................................................
    ! Clean up:

    deallocate(outIdInSrcData)
    deallocate(outIdBandData)
    deallocate(outIdInstData)
    deallocate(outIdSrcData)

    call release(outTab)
    call release(outSet)

    deallocate(inIdBandData)
    deallocate(inIdInstData)
    deallocate(inIdSrcData)

    call release(inTab)
    call release(inSet)

  end subroutine esrcfetch
  
  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: numInfoCols, i

    !-------------------------------------------------------------------
    ! Start

    inSetTabName = stringParameter('intable')
    outSetTabName = stringParameter('outtable')
    inIdSrcColName = stringParameter('inidcol')
    outIdSrcColName = stringParameter('outidcol')

    numInfoCols = parameterCount('infocols')
    allocate(infoColNames(numInfoCols))
    do i = 1, numInfoCols
      infoColNames(i) = stringParameter('infocols', i-1)
    end do

    flagSplitSources = booleanParameter('flagsplit')
    if (flagSplitSources) then
      flagColName = stringParameter('flagcol')
      flagBit = intParameter('flagbit')
    end if

    withUserIntNull = booleanParameter('withintnull')
    if (flagSplitSources) then
      userIntNullValue = intParameter('intnull')
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  logical function dataTypeIsInteger(dType)
  !---------------------------------------------------------------------
!*** put into dal_aux.

    implicit none

    integer(int32), intent(in) :: dType

    !-------------------------------------------------------------------
    ! Start

    dataTypeIsInteger = .false. ! default
    if (dType==INTEGER8 .or. dType==INTEGER16 .or. dType==INTEGER32)&
      dataTypeIsInteger = .true.

  end function dataTypeIsInteger

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module esrcfetch_mod
